from datetime import datetime
from io import BytesIO
from typing import TYPE_CHECKING, Any

import pandas as pd
from pandas.tseries.offsets import BMonthEnd
from wbfdm.models import InstrumentList
from wbreport.mixins import ReportMixin

if TYPE_CHECKING:
    from wbcore.contrib.authentication.models import User
    from wbreport.models import Report, ReportVersion


class ReportClass(ReportMixin):
    @classmethod
    def parse_parameters(cls, parameters: dict[str, str]) -> dict[str, Any]:
        return {
            "end": datetime.strptime(parameters["end"], "%Y-%m-%d").date(),
        }

    @classmethod
    def get_next_parameters(cls, parameters: dict[str, Any]) -> dict[str, Any]:
        parse_parameters = cls.parse_parameters(parameters)
        return {
            "end": datetime.strftime((parse_parameters["end"] + BMonthEnd(1)).date(), "%Y-%m-%d"),
        }

    @classmethod
    def get_version_date(cls, parameters) -> datetime.date:
        parameters = cls.parse_parameters(parameters)
        return parameters["end"]

    @classmethod
    def get_version_title(cls, report_title: str, parameters: dict[str, Any]) -> str:
        parameters = cls.parse_parameters(parameters)
        return f"{report_title} - {parameters['end']:%b %Y}"

    @classmethod
    def has_view_permission(cls, report: "Report", user: "User") -> bool:
        return user.has_perms(["wbreport.view_report", "wbportfolio.view_assetposition"])

    @classmethod
    def get_context(cls, version: "ReportVersion") -> dict[str, Any]:
        instrument_list: "InstrumentList" = version.report.content_object
        end_date = version.parameters.get("end")
        positions = []

        for instrument in instrument_list.instruments.all():
            if portfolio := instrument.primary_portfolio:
                for position in portfolio.assets.filter(date=end_date):
                    positions.append(
                        {
                            "portfolio": instrument.name,
                            "isin": position.underlying_instrument.isin,
                            "title": position.underlying_instrument.name_repr,
                            "instrument_type": position.underlying_instrument.instrument_type.short_name,
                            "weight": float(position.weighting),
                            "date": position.date.strftime("%Y-%m-%d"),
                        }
                    )

        return {"positions": positions}

    @classmethod
    def generate_file(cls, context: dict[str, Any]) -> BytesIO:
        stream = BytesIO()
        if positions := context.get("positions", None):
            df = pd.DataFrame(positions).sort_values(by=["weight"], ascending=False)
            writer = pd.ExcelWriter(stream, engine="xlsxwriter")
            for portfolio, dff in df.groupby("portfolio"):
                dff.to_excel(writer, sheet_name=portfolio[0:31], index=False)
            writer.save()
        return stream



import pytest
from faker import Faker

from wbportfolio.risk_management.backends.ucits_portfolio import (
    RuleBackend as UCITPortfolioRuleBackend,
)
from wbportfolio.tests.models.utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestUcitsRuleModel(PortfolioTestMixin):
    @pytest.fixture
    def ucits_backend(self, weekday, portfolio):
        return UCITPortfolioRuleBackend(weekday, portfolio, [])

    def testcheck_rule_1(self, weekday, portfolio, asset_position_factory, ucits_backend):
        #  Check No single asset can represent more than 10% of the fund's assets;
        asset_position_factory.create(date=weekday, weighting=0.05, portfolio=portfolio)
        asset_position_factory.create(date=weekday, weighting=0.05, portfolio=portfolio)
        a3 = asset_position_factory.create(date=weekday, weighting=0.15, portfolio=portfolio)

        res = list(ucits_backend.check_rule())
        assert len(res) == 1
        assert res[0].breached_object == a3.underlying_instrument

    def testcheck_rule_2(self, weekday, portfolio, asset_position_factory, ucits_backend):
        # Check that stock below 5% don't trigger rule
        asset_position_factory.create_batch(4, date=weekday, weighting=0.05, portfolio=portfolio)
        asset_position_factory.create_batch(20, date=weekday, weighting=0.04, portfolio=portfolio)

        res = list(ucits_backend.check_rule())
        assert len(res) == 0

    def testcheck_rule_3(self, weekday, portfolio, asset_position_factory, ucits_backend):
        # Check holdings of more than 5% cannot in aggregate exceed 40% of the fund's assets
        asset_position_factory.create_batch(20, date=weekday, weighting=0.05, portfolio=portfolio)
        res = list(ucits_backend.check_rule())
        assert len(res) == 20



import pytest


@pytest.mark.django_db
class TestCustodianModel:
    def test_init(self, custodian_factory):
        custodian = custodian_factory.create()
        assert custodian.id is not None

    def test_str(self, custodian_factory):
        custodian = custodian_factory.create()
        assert str(custodian) == f"{custodian.name}"



from wbcore import serializers

from wbportfolio.models.portfolio_cash_targets import PortfolioCashTarget
from wbportfolio.serializers.portfolios import PortfolioRepresentationSerializer


class PortfolioCashTargetModelSerializer(serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")

    class Meta:
        model = PortfolioCashTarget
        fields = (
            "id",
            "portfolio",
            "_portfolio",
            "valid_date",
            "min_target",
            "target",
            "max_target",
            "comment",
        )



# Import CRM Relevant Data
from .portfolio_relationships import PortfolioInstrumentPreferredClassificationThroughInlineModelAdmin
from .asset import AssetPositionModelAdmin
from .custodians import CustodianModelAdmin
from .products import ProductAdmin
from .indexes import IndexAdmin
from .product_groups import ProductGroupAdmin
from .portfolio import PortfolioModelAdmin
from .registers import RegisterModelAdmin
from .roles import PortfolioRoleAdmin
from .transactions import DividendAdmin, FeesAdmin, TradeAdmin
from .reconciliations import AccountReconciliationAdmin
from .orders import OrderProposalAdmin
from .rebalancing import RebalancingModelAdmin, RebalancerAdmin



from decimal import Decimal
from typing import Type

from django.db.models import Exists, OuterRef, Sum
from django.db.models.expressions import F
from django.db.models.functions import Coalesce
from django.db.models.query import QuerySet
from wbcore import serializers
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.serializers.serializers import Serializer
from wbcrm.models import Account
from wbhuman_resources.models.kpi import KPI, KPIHandler
from wbhuman_resources.serializers import KPIModelSerializer
from wbportfolio.models.transactions.claim import Claim


class NetNewMoneyKPISerializer(KPIModelSerializer):
    transaction_subtype = serializers.ChoiceField(
        default="all",
        choices=[("SUBSCRIPTION", "Only Subscriptions"), ("REDEMPTION", "Only Redemptions"), ("all", "All")],
    )
    nnm_from = serializers.ChoiceField(
        default="all",
        label="NNM from",
        choices=[("new_clients", "New clients"), ("existing_clients", "Existing clients"), ("all", "All")],
    )
    only_approved = serializers.BooleanField(
        default=False, label="Only Approved claims", help_text="Filter only approve claims"
    )

    creator = serializers.BooleanField(
        default=True, label="Creator of claim", help_text="NNM considered are related to the creator of claim"
    )
    claimant = serializers.BooleanField(
        default=True, label="Claimant of claim", help_text="NNM considered are related to the claimant"
    )
    in_charge_of_customer = serializers.BooleanField(
        default=True,
        label="In charge of customer",
        help_text="NNM considered are related to the persons in charge of customer",
    )

    def update(self, instance, validated_data):
        transaction_subtype = validated_data.get(
            "transaction_subtype",
            instance.additional_data["serializer_data"].get("transaction_subtype", "all"),
        )
        only_approved = validated_data.get(
            "only_approved",
            instance.additional_data["serializer_data"].get("only_approved", False),
        )
        nnm_from = validated_data.get(
            "nnm_from",
            instance.additional_data["serializer_data"].get("nnm_from", "all"),
        )

        creator = validated_data.get(
            "creator",
            instance.additional_data["serializer_data"].get("creator", True),
        )
        claimant = validated_data.get(
            "claimant",
            instance.additional_data["serializer_data"].get("claimant", True),
        )
        in_charge_of_customer = validated_data.get(
            "in_charge_of_customer",
            instance.additional_data["serializer_data"].get("in_charge_of_customer", True),
        )

        additional_data = instance.additional_data
        additional_data["serializer_data"]["only_approved"] = only_approved
        additional_data["serializer_data"]["transaction_subtype"] = transaction_subtype
        additional_data["serializer_data"]["nnm_from"] = nnm_from
        additional_data["serializer_data"]["creator"] = creator
        additional_data["serializer_data"]["claimant"] = claimant
        additional_data["serializer_data"]["in_charge_of_customer"] = in_charge_of_customer

        additional_data["list_data"] = instance.get_handler().get_list_data(additional_data["serializer_data"])
        validated_data["additional_data"] = additional_data

        return super().update(instance, validated_data)

    class Meta(KPIModelSerializer.Meta):
        fields = (
            *KPIModelSerializer.Meta.fields,
            "transaction_subtype",
            "nnm_from",
            "creator",
            "claimant",
            "in_charge_of_customer",
            "only_approved",
        )


class NetNewMoneyKPI(KPIHandler):
    def get_name(self) -> str:
        return "Net New Money"

    def get_serializer(self) -> Type[Serializer]:
        return NetNewMoneyKPISerializer

    def annotate_parameters(self, queryset: QuerySet[KPI]) -> QuerySet[KPI]:
        return queryset.annotate(
            transaction_subtype=F("additional_data__serializer_data__transaction_subtype"),
            nnm_from=F("additional_data__serializer_data__nnm_from"),
            creator=F("additional_data__serializer_data__creator"),
            claimant=F("additional_data__serializer_data__claimant"),
            in_charge_of_customer=F("additional_data__serializer_data__in_charge_of_customer"),
            only_approved=F("additional_data__serializer_data__only_approved"),
        )

    def get_list_data(self, serializer_data: dict) -> list[str]:
        return [
            f"Claim Area: {serializer_data['transaction_subtype']}",
            f"NNM from: {serializer_data['nnm_from']}",
            f"Creator: {serializer_data['creator']}",
            f"Claimant: {serializer_data['claimant']}",
            f"In charge of customer: {serializer_data['in_charge_of_customer']}",
            f"Only Approved: {serializer_data['only_approved']}",
        ]

    def get_display_grid(self) -> list[list[str]]:
        return [
            ["nnm_from"] * 3,
            ["only_approved"] * 3,
            ["transaction_subtype"] * 3,
            ["creator", "claimant", "in_charge_of_customer"],
        ]

    def evaluate(self, kpi: "KPI", evaluated_person, evaluation_date=None) -> int:
        serializer_data = kpi.additional_data.get("serializer_data")
        to_date = evaluation_date if evaluation_date else kpi.period.upper
        claims = (
            Claim.objects.exclude(status=Claim.Status.WITHDRAWN)
            .filter(date__gte=kpi.period.lower, date__lte=to_date, account__isnull=False)
            .annotate(
                existing_client=Exists(Claim.objects.filter(date__lte=kpi.period.lower, account=OuterRef("account")))
            )
        )

        if serializer_data.get("only_approved") is True:
            claims = claims.filter(status=Claim.Status.APPROVED)

        if (nnm_from := serializer_data.get("nnm_from")) and (nnm_from != "all"):
            if nnm_from == "new_clients":
                claims = claims.filter(existing_client=False)
            elif nnm_from == "existing_clients":
                claims = claims.filter(existing_client=True)

        # TODO This introduced duplicates and can't be removed easily with distinct("id") because of the aggregation
        accounts_ids = list(
            Account.get_managed_accounts_for_entry(evaluated_person.entry_ptr).values_list("id", flat=True)
        )
        for employer in evaluated_person.employers.all():
            accounts_ids.extend(list(Account.get_managed_accounts_for_entry(employer).values_list("id", flat=True)))
        accounts = Account.objects.filter(id__in=accounts_ids)
        claims = claims.filter(account__in=accounts)

        return claims.annotate(
            net_new_money=F("trade__price") * F("shares"),
            fx_rate=CurrencyFXRates.get_fx_rates_subquery("date", currency="product__currency", lookup_expr="exact"),
            net_new_money_usd=Coalesce(F("net_new_money") * F("fx_rate"), Decimal(0)),
        ).aggregate(s=Sum("net_new_money"))["s"] or Decimal(0)



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ESGMetricAggregationPortfolioPandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None



import factory

from wbportfolio.models import PortfolioRole


class ProductPortfolioRoleFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = PortfolioRole

    role_type = factory.Iterator([PortfolioRole.RoleType.PORTFOLIO_MANAGER, PortfolioRole.RoleType.ANALYST])
    person = factory.SubFactory("wbcore.contrib.authentication.factories.users.AuthenticatedPersonFactory")
    instrument = factory.SubFactory("wbportfolio.factories.ProductFactory")
    weighting = factory.Faker("pyfloat", min_value=0.1, max_value=1)


class ManagerPortfolioRoleFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = PortfolioRole

    role_type = factory.Iterator([PortfolioRole.RoleType.RISK_MANAGER, PortfolioRole.RoleType.MANAGER])
    person = factory.SubFactory("wbcore.contrib.directory.factories.entries.PersonFactory")



# Generated by Django 5.0.9 on 2024-10-16 12:15

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0067_assetposition_unique_asset_position"),
    ]

    operations = [
        migrations.AddField(
            model_name="trade",
            name="internal_trade",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="internal_subscription_redemption_trade",
                to="wbportfolio.trade",
            ),
        ),
        migrations.AddField(
            model_name="trade",
            name="marked_as_internal",
            field=models.BooleanField(
                default=False,
                help_text="If this is checked, then this subscription or redemption is considered internal and will not be considered in any AUM computation",
                verbose_name="Internal",
            ),
        ),
        migrations.AddConstraint(
            model_name="trade",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("marked_as_internal", False),
                    models.Q(
                        ("marked_as_internal", True), ("transaction_subtype__in", ["REDEMPTION", "SUBSCRIPTION"])
                    ),
                    _connector="OR",
                ),
                name="marked_as_internal_only_for_subred",
            ),
        ),
        migrations.AddConstraint(
            model_name="trade",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("internal_trade__isnull", True),
                    models.Q(
                        ("internal_trade__isnull", False), ("transaction_subtype__in", ["REDEMPTION", "SUBSCRIPTION"])
                    ),
                    _connector="OR",
                ),
                name="internal_trade_set_only_for_subred",
            ),
        ),
    ]



from datetime import date, timedelta
from decimal import Decimal

import pytest
from django.conf import settings
from dynamic_preferences.registries import global_preferences_registry
from faker import Faker
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.directory.factories import CompanyFactory, CustomerStatusFactory
from wbcore.contrib.directory.models import Company

from wbportfolio.contrib.company_portfolio.models import (
    CompanyPortfolioData,
    get_client_customer_status,
    get_lost_client_customer_status,
    get_returning_client_customer_status,
    get_tpm_customer_status,
)

fake = Faker()


def custom_get_potential(company_portfolio_data, val_date: date):
    return val_date


def custom_get_tiering(company_portfolio_data, aum: Decimal):
    return aum


@pytest.mark.django_db
class TestCompanyPortfolioDataUpdater:
    @pytest.fixture
    def company_portfolio_data(self):
        company = CompanyFactory.create()
        usd = Currency.objects.get_or_create(key="USD")[0]
        global_preferences = global_preferences_registry.manager()
        global_preferences["wbportfolio__tpm_customer_status"] = CustomerStatusFactory.create(title="TPM")
        global_preferences["wbportfolio__client_customer_status"] = CustomerStatusFactory.create(title="Client")
        return CompanyPortfolioData.objects.get_or_create(
            company=company, defaults={"potential_currency": usd, "assets_under_management_currency": usd}
        )[0]

    def test_get_assets_under_management_usd(self, company_portfolio_data, claim):
        claim.account.owner = company_portfolio_data.company
        claim.account.save()
        claim.status = "APPROVED"
        claim.save()
        assert company_portfolio_data.get_assets_under_management_usd(claim.date - timedelta(days=1)) == Decimal(0)
        assert (
            company_portfolio_data.get_assets_under_management_usd(claim.date)
            == claim.shares * claim.product.valuations.get(date=claim.date).net_value
        )

    def test__get_default_potential(
        self, company_portfolio_data, weekday, currency_fx_rates_factory, asset_allocation_factory
    ):
        currency_fx_rates_factory.create(
            currency=company_portfolio_data.assets_under_management_currency, date=weekday, value=Decimal(1)
        )

        initial_aum = Decimal(1000)
        invested_aum = Decimal(100)
        company_portfolio_data.assets_under_management = initial_aum
        company_portfolio_data.invested_assets_under_management_usd = invested_aum

        asset_allocation_1 = asset_allocation_factory.create(
            company=company_portfolio_data.company, percent=Decimal(0.8), max_investment=Decimal(0.2)
        )
        asset_allocation_2 = asset_allocation_factory.create(
            company=company_portfolio_data.company, percent=Decimal(0.2), max_investment=Decimal(0.3)
        )
        asset_allocation_1.refresh_from_db()
        asset_allocation_2.refresh_from_db()
        assert (
            company_portfolio_data._get_default_potential(weekday)
            == initial_aum * asset_allocation_1.percent * asset_allocation_1.max_investment
            + initial_aum * asset_allocation_2.percent * asset_allocation_2.max_investment
            - invested_aum
        )

    def test__get_default_tiering_for_client(self, company_portfolio_data):
        company_portfolio_data.company.customer_status = get_client_customer_status()
        company_portfolio_data.company.save()

        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(
            min_value=10, max_value=100
        ) / Decimal(100)
        assert company_portfolio_data._get_default_tiering(Decimal(1)) == Company.Tiering.ONE

        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(
            min_value=5, max_value=10
        ) / Decimal(100)
        assert company_portfolio_data._get_default_tiering(Decimal(1)) == Company.Tiering.TWO

        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(
            min_value=2, max_value=5
        ) / Decimal(100)
        assert company_portfolio_data._get_default_tiering(Decimal(1)) == Company.Tiering.THREE

        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(
            min_value=1, max_value=2
        ) / Decimal(100)
        assert company_portfolio_data._get_default_tiering(Decimal(1)) == Company.Tiering.FOUR

        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(max_value=1) / Decimal(100)
        assert company_portfolio_data._get_default_tiering(Decimal(1)) == Company.Tiering.FIVE

    def test_get_customer_status(self, company_portfolio_data):
        # without AUM, the method should return the company customer status
        assert company_portfolio_data.get_customer_status() == company_portfolio_data.company.customer_status

        # add unvalid AUM
        company_portfolio_data.invested_assets_under_management_usd = -10
        assert (
            company_portfolio_data.get_customer_status() == company_portfolio_data.company.customer_status
        )  # unvalid aum will return as no aum

        # add positive AUM
        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(positive=True)
        assert (
            company_portfolio_data.get_customer_status() == get_client_customer_status()
        )  # unvalid aum will return as no aum

        # ensure TPM are not considered as client even if they have AUM
        company_portfolio_data.company.customer_status = get_tpm_customer_status()
        company_portfolio_data.company.save()
        assert company_portfolio_data.get_customer_status() == get_tpm_customer_status()

    def test_get_lost_customer_status(self, company_portfolio_data):
        company_portfolio_data.company.customer_status = get_client_customer_status()
        company_portfolio_data.company.save()

        company_portfolio_data.invested_assets_under_management_usd = None
        assert company_portfolio_data.get_customer_status() == get_lost_client_customer_status()

    def test_returning_customer_status(self, company_portfolio_data):
        company_portfolio_data.company.customer_status = get_lost_client_customer_status()
        company_portfolio_data.company.save()

        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(positive=True)
        assert company_portfolio_data.get_customer_status() == get_returning_client_customer_status()

    def test_staying_returning_customer_status(self, company_portfolio_data):
        company_portfolio_data.company.customer_status = get_returning_client_customer_status()
        company_portfolio_data.company.save()

        company_portfolio_data.invested_assets_under_management_usd = fake.pydecimal(positive=True)
        assert company_portfolio_data.get_customer_status() == get_returning_client_customer_status()

    def test_get_potential_from_custom_method(self, company_portfolio_data, weekday):
        settings.PORTFOLIO_COMPANY_DATA_POTENTIAL_METHOD = (
            "wbportfolio.contrib.company_portfolio.tests.test_models.custom_get_potential"
        )
        assert company_portfolio_data.get_potential(weekday) == weekday

    @pytest.mark.parametrize("total_aum", [fake.pydecimal()])
    def test_get_tiering_from_custom_method(self, company_portfolio_data, total_aum):
        settings.PORTFOLIO_COMPANY_DATA_TIERING_METHOD = (
            "wbportfolio.contrib.company_portfolio.tests.test_models.custom_get_potential"
        )
        assert company_portfolio_data.get_tiering(total_aum) == total_aum



from django.contrib import admin

from wbportfolio.models import AccountReconciliation, AccountReconciliationLine


class AccountReconciliationLineInline(admin.TabularInline):
    model = AccountReconciliationLine
    fields = ("product", "price", "price_date", "shares", "shares_external")
    extra = 0


@admin.register(AccountReconciliation)
class AccountReconciliationAdmin(admin.ModelAdmin):
    list_display = ("reconciliation_date", "account", "creator")
    inlines = [AccountReconciliationLineInline]



from wbcore.metadata.configs.titles import TitleViewConfig


class RegisterTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Registers"

    def get_instance_title(self):
        return "Register: {{register_reference}}"

    def get_create_title(self):
        return "New Register"



# Generated by Django 5.0.14 on 2025-07-10 09:00

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0079_alter_trade_drift_factor'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assetposition',
            name='weighting',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'), help_text='The Weight of the Asset on the price date of the Asset.', max_digits=9, verbose_name='Weight'),
        ),
    ]



import datetime
import logging
import re
from io import BytesIO

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import extract_exchange_ticker

logger = logging.getLogger("importers.parsers.jp_morgan.strategy")


def file_name_parse(file_name):
    dates = re.findall("([0-9]{8})", file_name)

    assert len(dates) == 1, "Not exactly 1 date found in the filename"

    return {"valuation_date": datetime.datetime.strptime(dates[0], "%Y%m%d").date()}


def parse(import_source):
    data = list()
    prices = list()
    df_dict = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl", sheet_name=None)
    for sheet_name, df in df_dict.items():
        xx, yy = np.where(df == "Ticker")
        if len(xx) == 1 and len(yy) == 1:
            df_info = df.iloc[: xx[0] - 1, :].transpose()
            df_info = df_info.rename(columns=df_info.iloc[0]).drop(df_info.index[0]).dropna(how="all")

            strategy_ticker = df_info.loc[:, df_info.columns.str.contains("Ticker")].iloc[0, 0]
            strategy_currency = df_info.loc[:, df_info.columns.str.contains("CCY")].iloc[0, 0]
            valuation_date = df_info.loc[:, df_info.columns.str.contains("Date")].iloc[0, 0]
            valuation_date = datetime.datetime.strptime(valuation_date, "%Y-%m-%d")
            strategy_close = df_info.loc[:, df_info.columns.str.contains("Level")].iloc[0, 0]

            df_positions = df.iloc[xx[0] :, yy[0] :]
            df_positions = (
                df_positions.rename(columns=df_positions.iloc[0]).drop(df_positions.index[0]).dropna(how="all")
            )
            df_positions = df_positions.rename(
                columns={
                    df_positions.columns[df_positions.columns.str.contains("Forex")][0]: "initial_currency_fx_rate",
                    df_positions.columns[df_positions.columns.str.contains("Currency")][0]: "currency__key",
                    df_positions.columns[df_positions.columns.str.contains("Weight")][0]: "weighting",
                    df_positions.columns[df_positions.columns.str.contains("Level")][0]: "initial_price",
                    df_positions.columns[df_positions.columns.str.contains("Ticker")][0]: "underlying_quote__ticker",
                    df_positions.columns[df_positions.columns.str.contains("ISIN")][0]: "underlying_quote__isin",
                    df_positions.columns[df_positions.columns.str.contains("Name")][0]: "underlying_quote__name",
                }
            )
            for position in df_positions.to_dict("records"):
                ticker, exchange = extract_exchange_ticker(position["underlying_quote__ticker"])
                if exchange:
                    exchange = {"bbg_exchange_codes": exchange}
                data.append(
                    {
                        "underlying_quote": {
                            "instrument_type": "equity",
                            "ticker": ticker,
                            "isin": position["underlying_quote__isin"],
                            "name": position["underlying_quote__name"],
                            "currency__key": position["currency__key"],
                            "exchange": exchange,
                        },
                        "portfolio": {
                            "instrument_type": "index",
                            "ticker": strategy_ticker,
                            "currency__key": strategy_currency,
                        },
                        "is_estimated": False,
                        "exchange": exchange,
                        "asset_type": "equity",
                        "currency__key": position["currency__key"],
                        "initial_currency_fx_rate": round(position["initial_currency_fx_rate"], 14),
                        "weighting": round(position["weighting"], 8),
                        "initial_price": round(position["initial_price"], 6),
                        "date": valuation_date.strftime("%Y-%m-%d"),
                    }
                )
            prices.append(
                {
                    "instrument": {
                        "instrument_type": "index",
                        "ticker": strategy_ticker,
                        "currency__key": strategy_currency,
                    },
                    "date": valuation_date.strftime("%Y-%m-%d"),
                    "net_value": strategy_close,
                }
            )
    return {"data": data, "prices": prices}



from wbcore import viewsets
from wbfdm.viewsets.configs.buttons.instruments import InstrumentButtonViewConfig

from wbportfolio.models import ProductGroup
from wbportfolio.serializers import (
    ProductGroupModelSerializer,
    ProductGroupRepresentationSerializer,
)

from .configs import (
    ProductGroupDisplayConfig,
    ProductGroupEndpointConfig,
    ProductGroupTitleConfig,
)


class ProductGroupRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:product_group"
    serializer_class = ProductGroupRepresentationSerializer

    search_fields = ("computed_str",)

    queryset = ProductGroup.active_objects.all()


class ProductGroupModelViewSet(viewsets.ModelViewSet):
    queryset = ProductGroup.objects.all()
    serializer_class = ProductGroupModelSerializer

    title_config_class = ProductGroupTitleConfig
    endpoint_config_class = ProductGroupEndpointConfig
    display_config_class = ProductGroupDisplayConfig
    button_config_class = InstrumentButtonViewConfig

    ordering_fields = (
        "name",
        "identifier",
        "umbrella",
        "type",
        "category",
        "management_company",
        "depositary",
        "transfer_agent",
        "administrator",
        "investment_manager",
        "auditor",
        "paying_agent",
    )
    ordering = ("identifier",)
    search_fields = ("name", "identifier", "umbrella", "category", "type")

    filterset_fields = {
        "identifier": ["exact"],
        "name": ["exact"],
        "umbrella": ["exact"],
        "type": ["exact"],
        "category": ["exact"],
        "management_company": ["exact"],
        "depositary": ["exact"],
        "transfer_agent": ["exact"],
        "administrator": ["exact"],
        "investment_manager": ["exact"],
        "auditor": ["exact"],
        "paying_agent": ["exact"],
    }

    def get_serializer_class(self):
        return ProductGroupModelSerializer

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related(
                "management_company",
                "depositary",
                "transfer_agent",
                "administrator",
                "investment_manager",
                "auditor",
                "paying_agent",
                "instrument_ptr",
            )
            .prefetch_related("representants")
        )



from decimal import Decimal

from django.db import models
from wbcore.models import WBModel


class PortfolioCashTarget(WBModel):
    """This model stores cash targets for a given portfolio"""

    portfolio = models.ForeignKey(
        to="wbportfolio.Portfolio",
        related_name="cash_targets",
        on_delete=models.CASCADE,
    )

    valid_date = models.DateField()
    min_target = models.DecimalField(max_digits=5, decimal_places=4, null=True, blank=True)
    target = models.DecimalField(max_digits=5, decimal_places=4, default=Decimal(0))
    max_target = models.DecimalField(max_digits=5, decimal_places=4, null=True, blank=True)
    comment = models.TextField(default="", blank=True)

    def __str__(self) -> str:
        return f"{self.portfolio}: {self.target:.2%} ({self.valid_date:%d.%m.%Y})"

    class Meta:
        verbose_name = "Portfolio Cash Target"
        verbose_name_plural = "Portfolio Cash Targets"
        constraints = [
            models.UniqueConstraint(fields=["portfolio", "valid_date"], name="unique_portfolio_valid_date"),
        ]

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfoliocashtarget"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:portfoliocashtarget-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{portfolio}}: {{target}} ({{valid_date}})"



# Generated by Django 5.0.14 on 2025-08-11 14:01

import django.db.models.deletion
import django_fsm
from django.db import migrations, models

def migrate_approve_to_apply(apps, schema_editor):
    OrderProposal = apps.get_model("wbportfolio", "OrderProposal")
    OrderProposal.objects.filter(status="APPROVED").update(status="APPLIED")


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0086_orderproposal_total_cash_weight'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='order_routing_custodian_adapter',
            field=models.CharField(blank=True, help_text='The dotted path to the order routing custodian adapter', max_length=1024, null=True, verbose_name='Order Routing Custodian Adapter'),
        ),
        migrations.AddField(
            model_name='order',
            name='execution_comment',
            field=models.TextField(blank=True, default='', verbose_name='Execution Comment'),
        ),
        migrations.AddField(
            model_name='order',
            name='execution_confirmed',
            field=models.BooleanField(default=False, verbose_name='Execution Confirmed'),
        ),
        migrations.AddField(
            model_name='orderproposal',
            name='approver',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT,
                                    related_name='approver_order_proposals', to='directory.person',
                                    verbose_name='Approver'),
        ),
        migrations.AddField(
            model_name='orderproposal',
            name='execution_comment',
            field=models.CharField(blank=True, default='', verbose_name='Execution Comment'),
        ),
        migrations.AddField(
            model_name='orderproposal',
            name='execution_status',
            field=models.CharField(blank=True, choices=[('PENDING', 'Pending'), ('IN_DRAFT', 'In Draft'),
                                                        ('COMPLETED', 'Completed'), ('CANCELLED', 'Cancelled'),
                                                        ('REJECTED', 'Rejected'), ('FAILED', 'Failed'),
                                                        ('UNKNOWN', 'Unknown')], default='',
                                   verbose_name='Execution Status'),
        ),
        migrations.AddField(
            model_name='orderproposal',
            name='execution_status_detail',
            field=models.CharField(blank=True, default='', verbose_name='Execution Status Detail'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='status',
            field=django_fsm.FSMField(choices=[('DRAFT', 'Draft'), ('SUBMIT', 'Pending'), ('APPROVED', 'Approved'),
                                               ('PENDING_EXECUTION', 'Pending Execution'), ('EXECUTED', 'Executed'),
                                               ('DENIED', 'Denied'), ('FAILED', 'Failed')], default='DRAFT',
                                      max_length=50, verbose_name='Status'),
        ),
        migrations.AddField(
            model_name='index',
            name='order_routing_custodian_adapter',
            field=models.CharField(blank=True, help_text='The dotted path to the order routing custodian adapter',
                                   max_length=1024, null=True, verbose_name='Order Routing Custodian Adapter'),
        ),
        migrations.AddField(
            model_name='productgroup',
            name='order_routing_custodian_adapter',
            field=models.CharField(blank=True, help_text='The dotted path to the order routing custodian adapter',
                                   max_length=1024, null=True, verbose_name='Order Routing Custodian Adapter'),
        ),
        migrations.RenameField(
            model_name='rebalancer',
            old_name='approve_order_proposal_automatically',
            new_name='apply_order_proposal_automatically',
        ),
        migrations.RunPython(migrate_approve_to_apply),
        migrations.AlterField(
            model_name='orderproposal',
            name='status',
            field=django_fsm.FSMField(
                choices=[('DRAFT', 'Draft'), ('PENDING', 'Pending'), ('APPROVED', 'Approved'), ('DENIED', 'Denied'),
                         ('APPLIED', 'Applied'), ('EXECUTION', 'Execution'), ('FAILED', 'Failed')], default='DRAFT',
                max_length=50, verbose_name='Status'),
        ),
    ]



import factory
from faker import Faker

from wbportfolio.factories.portfolios import PortfolioFactory
from wbportfolio.models import DailyPortfolioCashFlow

fake = Faker()


class DailyPortfolioCashFlowFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = DailyPortfolioCashFlow

    value_date = factory.Faker("date")
    portfolio = factory.SubFactory(PortfolioFactory)

    total_assets = factory.Faker("pydecimal", left_digits=7, right_digits=4, min_value=0.01)
    cash = factory.Faker("pydecimal", left_digits=7, right_digits=4, min_value=0.01)
    cash_flow_forecast = factory.LazyAttribute(lambda o: fake.pydecimal(min_value=0.01, max_value=o.total_assets))

    pending = factory.Faker("pybool")



# Generated by Django 5.0.8 on 2024-08-13 13:48

import django.db.models.manager
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0063_accountreconciliation_accountreconciliationline_and_more"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="portfolio",
            managers=[
                ("tracked_objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name="portfolio",
            name="is_tracked",
            field=models.BooleanField(
                default=True,
                help_text="True if the internal updating mechanism (e.g., Propagation, Synchronization etc...) needs to apply to this portfolio",
            ),
        ),
        migrations.AlterField(
            model_name="portfolio",
            name="is_invested",
            field=models.BooleanField(default=False, help_text="True if the portfolio AUM counts"),
        ),
        migrations.AlterField(
            model_name="portfolio",
            name="is_manageable",
            field=models.BooleanField(
                default=False,
                help_text="True if the portfolio can be manually modified (e.g. Trade proposal be submitted or total weight recomputed)",
            ),
        ),
    ]



from decimal import Decimal

from django.contrib.contenttypes.models import ContentType
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.serializers import (
    CompanyRepresentationSerializer,
    EntryRepresentationSerializer,
)
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer
from wbfdm.models import Instrument
from wbfdm.serializers.instruments.instruments import InstrumentModelSerializer

from wbportfolio.models import Portfolio, Product
from wbportfolio.serializers.portfolios import PortfolioRepresentationSerializer

from .product_group import ProductGroupRepresentationSerializer


class ProductRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:product-detail")
    bank_name = wb_serializers.CharField(read_only=True)

    class Meta:
        model = Product
        fields = ("id", "computed_str", "isin", "name", "bank_name", "_detail")


class ProductCustomerRepresentationSerializer(ProductRepresentationSerializer):
    label_key = "{{name}} - {{isin}} ({{bank_name}})"

    class Meta:
        model = Product
        fields = ("id", "isin", "name", "bank_name")


class ProductUnlinkedRepresentationSerializer(ProductRepresentationSerializer):
    class Meta(ProductRepresentationSerializer.Meta):
        model = Product
        fields = ("id", "isin", "name")


class ProductListModelSerializer(
    InstrumentModelSerializer,
):
    _bank = CompanyRepresentationSerializer(source="bank")
    _parent = ProductGroupRepresentationSerializer(source="parent")
    _white_label_customers = EntryRepresentationSerializer(source="white_label_customers", many=True)
    current_total_issuer_fees = wb_serializers.DecimalField(
        default=Decimal(0), read_only=True, max_digits=6, decimal_places=4, label="Issue Fees"
    )
    current_bank_fees = wb_serializers.DecimalField(
        default=Decimal(0), read_only=True, max_digits=6, decimal_places=4, label="Bank Fees"
    )
    current_management_fees = wb_serializers.DecimalField(
        default=Decimal(0), read_only=True, max_digits=6, decimal_places=4, label="Management Fees"
    )
    current_performance_fees = wb_serializers.DecimalField(
        default=Decimal(0), read_only=True, max_digits=6, decimal_places=4, label="Performance Fees"
    )

    # AUM based info
    net_value = wb_serializers.FloatField(label="Net Value", required=False, read_only=True)
    assets_under_management = wb_serializers.FloatField(
        label="Assets under Management", required=False, default=0, read_only=True
    )
    assets_under_management_usd = wb_serializers.FloatField(
        label="Assets under Management in USD", required=False, default=0, read_only=True
    )
    sum_shares = wb_serializers.FloatField(default=0, read_only=True)
    is_white_label = wb_serializers.BooleanField(read_only=True)

    # NNM annotations
    nnm_weekly = wb_serializers.FloatField(label="NNM Weekly", required=False, default=0, read_only=True)
    nnm_monthly = wb_serializers.FloatField(label="NNM Monthly", required=False, default=0, read_only=True)
    nnm_year_to_date = wb_serializers.FloatField(label="NNM YTD", required=False, default=0, read_only=True)
    nnm_yearly = wb_serializers.FloatField(label="NNM YEARLY", required=False, default=0, read_only=True)
    # general info
    is_invested = wb_serializers.BooleanField(default=True, read_only=True)

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        additional_resources = dict()
        additional_resources["peers"] = (
            reverse("wbfdm:instrument-relatedinstrument-list", args=[instance.id], request=request)
            + "?related_type=PEER"
        )
        additional_resources["benchmarks"] = (
            reverse("wbfdm:instrument-relatedinstrument-list", args=[instance.id], request=request)
            + "?related_type=BENCHMARK"
        )
        return additional_resources

    class Meta(InstrumentModelSerializer.Meta):
        decorators = {
            "share_price": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "net_value": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "assets_under_management": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "assets_under_management_usd": wb_serializers.decorator(decorator_type="text", position="left", value="$"),
            "nnm_weekly": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "nnm_monthly": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "nnm_year_to_date": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "nnm_yearly": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
        }
        percent_fields = [
            "current_bank_fees",
            "current_management_fees",
            "current_performance_fees",
            "current_total_issuer_fees",
        ]
        model = Product
        read_only_fields = [
            "share_price",
            "initial_high_water_mark",
            "issue_price",
            "last_valuation_date",
            "last_price_date",
            *InstrumentModelSerializer.Meta.read_only_fields,
        ]
        fields = (
            "id",
            "name",
            "name_repr",
            "bank",
            "_bank",
            "parent",
            "_parent",
            "description",
            "share_price",
            "initial_high_water_mark",
            "current_total_issuer_fees",
            "current_management_fees",
            "current_performance_fees",
            "current_bank_fees",
            "currency",
            "_currency",
            "isin",
            "ticker",
            "net_value",
            "sum_shares",
            "assets_under_management",
            "assets_under_management_usd",
            "nnm_weekly",
            "nnm_monthly",
            "nnm_year_to_date",
            "nnm_yearly",
            "is_white_label",
            "last_valuation_date",
            "last_price_date",
            "is_invested",
            "white_label_customers",
            "_white_label_customers",
            "_additional_resources",
            "_classifications",
            "classifications",
        )


class ProductModelSerializer(ProductListModelSerializer):
    id_repr = wb_serializers.CharField(source="id", read_only=True, label="ID")

    _jurisdiction = GeographyRepresentationSerializer(source="jurisdiction", filter_params={"level": 1})

    is_next_factsheet_available = wb_serializers.BooleanField(default=False, read_only=True)

    _portfolios = PortfolioRepresentationSerializer(source="portfolios", many=True)

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        additional_resources = dict()

        if user.profile.is_internal or user.is_superuser:
            additional_resources["aum"] = (
                f'{reverse("wbportfolio:aumtable-list", args=[], request=request)}?product={instance.id}&group_by=ACCOUNT'
            )
            additional_resources["claims"] = reverse(
                "wbportfolio:product-claim-list",
                args=[instance.id],
                request=request,
            )

            if portfolio := instance.portfolio:
                additional_resources["preferredclassification"] = reverse(
                    "wbportfolio:portfolio-preferredclassification-list",
                    args=[portfolio.id],
                    request=request,
                )
                product_ct = ContentType.objects.get_for_model(Product)
                instrument_ct = ContentType.objects.get_for_model(Instrument)
                portfolio_cd = ContentType.objects.get_for_model(Portfolio)
                additional_resources["risk_rules"] = (
                    reverse("wbcompliance:riskrule-list", args=[], request=request)
                    + f"?checked_object_relationships=[({product_ct.id},{instance.id}),({instrument_ct.id},{instance.id}),({portfolio_cd.id},{portfolio.id})]"
                )
                additional_resources["risk_incidents"] = (
                    reverse("wbcompliance:riskincident-list", args=[], request=request)
                    + f"?checked_object_relationships=[({product_ct.id},{instance.id}),({instrument_ct.id},{instance.id}),({portfolio_cd.id},{portfolio.id})]"
                )
        return additional_resources

    class Meta(ProductListModelSerializer.Meta):
        fields = (
            "id_repr",
            "bank",
            "_bank",
            "share_price",
            "initial_high_water_mark",
            "current_bank_fees",
            "current_management_fees",
            "current_performance_fees",
            "current_total_issuer_fees",
            "white_label_customers",
            "_white_label_customers",
            "type_of_return",
            "asset_class",
            "legal_structure",
            "investment_index",
            "liquidity",
            "risk_scale",
            "jurisdiction",
            "_jurisdiction",
            "is_next_factsheet_available",
            "external_webpage",
            "issue_price",
            "is_invested",
            "net_value",
            "last_valuation_date",
            "assets_under_management",
            "assets_under_management_usd",
            "nnm_weekly",
            "nnm_monthly",
            "nnm_year_to_date",
            "nnm_yearly",
            "is_white_label",
            "parent",
            "_parent",
            "_portfolios",
            "portfolios",
        ) + InstrumentModelSerializer.Meta.fields


class ProductCustomerModelSerializer(ProductListModelSerializer):
    bank_repr = wb_serializers.CharField(read_only=True)
    currency_symbol = wb_serializers.CharField(read_only=True)

    class Meta:
        model = Product
        decorators = {
            "net_value": wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}"),
        }
        fields = (
            "id",
            "name",
            "isin",
            "bank_repr",
            "currency_symbol",
            "ticker",
            "net_value",
        )


class ProductFeesModelSerializer(wb_serializers.ModelSerializer):
    _currency = CurrencyRepresentationSerializer(source="currency")

    sum_management_fees = wb_serializers.FloatField(
        label="Bank Fees",
        required=False,
        read_only=True,
        decorators=[{"position": "left", "value": "{{_currency.symbol}}"}],
    )
    sum_management_fees_usd = wb_serializers.FloatField(
        label="USD",
        required=False,
        read_only=True,
        decorators=[{"position": "left", "value": "$"}],
    )
    sum_performance_fees_net = wb_serializers.FloatField(
        label="Bank Fees",
        required=False,
        read_only=True,
        decorators=[{"position": "left", "value": "{{_currency.symbol}}"}],
    )
    sum_performance_fees_net_usd = wb_serializers.FloatField(
        label="USD",
        required=False,
        read_only=True,
        decorators=[{"position": "left", "value": "$"}],
    )
    sum_total = wb_serializers.FloatField(
        label="USD",
        required=False,
        read_only=True,
        decorators=[{"position": "left", "value": "{{_currency.symbol}}"}],
    )
    sum_total_usd = wb_serializers.FloatField(
        label="USD",
        required=False,
        read_only=True,
        decorators=[{"position": "left", "value": "$"}],
    )

    class Meta:
        model = Product
        fields = (
            "id",
            "computed_str",
            "isin",
            "ticker",
            "currency",
            "_currency",
            "sum_management_fees",
            "sum_management_fees_usd",
            "sum_performance_fees_net",
            "sum_performance_fees_net_usd",
            "sum_total",
            "sum_total_usd",
        )



import re

from fuzzywuzzy import process as fuzzywuzzy_process
from fuzzywuzzy import utils


def string_matching(s1, s2):
    best_result = ("", 0)
    for _s1 in s1.split(" "):
        if _s1 in ["&"]:
            continue
        # validate query
        if utils.full_process(_s1):
            matches = fuzzywuzzy_process.extract(_s1, s2)
            if matches[0][1] > best_result[1]:
                best_result = matches[0]

    return best_result


def re_ric(input: str):
    return set(re.findall(r"([A-Za-z0-9]{1,6}\.[A-Z]{1,2})(?:.)(?<!\.)", input))


def re_bloomberg(input: str):
    return set(re.findall(r"[A-Z]{2,5}(?:\-[A-Z]{2})?", input))


def re_isin(input: str):
    return set(re.findall(r"[A-Z]{2}[A-Z0-9]{9}[0-9]", input))



from datetime import date
from decimal import Decimal

from django.contrib import admin
from django.db import models
from django.db.models import (
    Sum,
)
from ordered_model.models import OrderedModel
from wbcore.contrib.io.mixins import ImportMixin

from wbportfolio.import_export.handlers.orders import OrderImportHandler
from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.transactions.transactions import TransactionMixin


class Order(TransactionMixin, ImportMixin, OrderedModel, models.Model):
    import_export_handler_class = OrderImportHandler

    ORDER_WEIGHTING_PRECISION = (
        8  # we need to match the asset position weighting. Skfolio advices using a even smaller number (5)
    )
    currency = None

    class Type(models.TextChoices):
        REBALANCE = "REBALANCE", "Rebalance"
        DECREASE = "DECREASE", "Decrease"
        INCREASE = "INCREASE", "Increase"
        BUY = "BUY", "Buy"
        SELL = "SELL", "Sell"
        NO_CHANGE = "NO_CHANGE", "No Change"  # default transaction subtype if weighing is 0

    order_type = models.CharField(max_length=32, default=Type.BUY, choices=Type.choices, verbose_name="Trade Type")
    shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        default=Decimal("0.0"),
        help_text="The number of shares that were traded.",
        verbose_name="Shares",
    )
    desired_target_weight = models.DecimalField(
        max_digits=9,
        decimal_places=ORDER_WEIGHTING_PRECISION,
        default=Decimal(0),
        help_text="Desired Target Weight (for compliance and audit)",
        verbose_name="Desired Target Weight",
    )
    weighting = models.DecimalField(
        max_digits=9,
        decimal_places=ORDER_WEIGHTING_PRECISION,
        default=Decimal(0),
        help_text="The weight to be multiplied against the target",
        verbose_name="Weight",
    )
    order_proposal = models.ForeignKey(
        to="wbportfolio.OrderProposal",
        related_name="orders",
        on_delete=models.CASCADE,
        help_text="The Order Proposal this trade is coming from",
    )
    daily_return = models.DecimalField(
        max_digits=ORDER_WEIGHTING_PRECISION * 2
        + 3,  # we don't expect any drift factor to be in the order of magnitude greater than 1000
        decimal_places=ORDER_WEIGHTING_PRECISION
        * 2,  # we need a higher precision for this factor to avoid float inprecision
        default=Decimal(0.0),
        verbose_name="Daily Return",
        help_text="The Ex-Post daily return",
    )

    execution_confirmed = models.BooleanField(default=False, verbose_name="Execution Confirmed")
    execution_comment = models.TextField(default="", blank=True, verbose_name="Execution Comment")

    def submit(self, by=None, description=None, portfolio_total_asset_value=None, **kwargs):
        warnings = []

        # if shares is defined and the underlying instrument defines a round lot size different than 1 and exchange allows its application, we round the share accordingly
        if self.order_proposal and not self.portfolio.only_weighting:
            shares = self.order_proposal.get_round_lot_size(self.shares, self.underlying_instrument)
            if shares != self.shares:
                warnings.append(
                    f"{self.underlying_instrument.computed_str} has a round lot size of  {self.underlying_instrument.round_lot_size}: shares were rounded from {self.shares} to {shares}"
                )
            shares = round(shares)  # ensure fractional shares are converted into integer
            # we need to recompute the delta weight has we changed the number of shares
            if shares != self.shares:
                self.shares = shares
                if portfolio_total_asset_value:
                    self.weighting = self.shares * self.price * self.currency_fx_rate / portfolio_total_asset_value
        if self.shares and abs(self.total_value_fx_portfolio) < self.order_proposal.min_order_value:
            warnings.append(
                f"Total Value for order {self.underlying_instrument.computed_str} ({self.total_value_fx_portfolio}) is bellow the allowed Minimum Order Value ({self.order_proposal.min_order_value})"
            )
            self.shares = Decimal("0")
            self.weighting = Decimal("0")
        if not self.price:
            warnings.append(f"No price for {self.underlying_instrument.computed_str}")
        if (
            not self.underlying_instrument.is_cash
            and not self.underlying_instrument.is_cash_equivalent
            and self._target_weight < -1e-8
        ):  # any value below -1e8 will be considered zero
            warnings.append(f"Negative target weight for {self.underlying_instrument.computed_str}")
        self.desired_target_weight = self._target_weight
        return warnings

    @property
    def product(self):
        from wbportfolio.models.products import Product

        try:
            return Product.objects.get(id=self.underlying_instrument.id)
        except Product.DoesNotExist:
            return None

    @property
    @admin.display(description="Last Effective Date")
    def _last_effective_date(self) -> date:
        if hasattr(self, "last_effective_date"):
            return self.last_effective_date
        elif (
            assets := AssetPosition.unannotated_objects.filter(
                date__lt=self.value_date,
                portfolio=self.portfolio,
            )
        ).exists():
            return assets.latest("date").date

    @property
    @admin.display(description="Effective Weight")
    def _previous_weight(self) -> Decimal:
        if hasattr(self, "previous_weight"):
            return self.previous_weight
        return AssetPosition.unannotated_objects.filter(
            underlying_quote=self.underlying_instrument,
            date=self._last_effective_date,
            portfolio=self.portfolio,
        ).aggregate(s=Sum("weighting"))["s"] or Decimal(0)

    @property
    @admin.display(description="Effective Weight")
    def _effective_weight(self) -> Decimal:
        if hasattr(self, "effective_weight"):
            return self.effective_weight
        return self.order_proposal.get_orders().get(id=self.id).effective_weight

    @property
    @admin.display(description="Effective Shares")
    def _effective_shares(self) -> Decimal:
        return getattr(
            self,
            "effective_shares",
            AssetPosition.objects.filter(
                underlying_quote=self.underlying_instrument,
                date=self.value_date,
                portfolio=self.portfolio,
            ).aggregate(s=Sum("shares"))["s"]
            or Decimal(0),
        )

    @property
    @admin.display(description="Target Weight")
    def _target_weight(self) -> Decimal:
        return getattr(
            self, "target_weight", round(self._effective_weight + self.weighting, self.ORDER_WEIGHTING_PRECISION)
        )

    @property
    @admin.display(description="Target Shares")
    def _target_shares(self) -> Decimal:
        return getattr(self, "target_shares", self._effective_shares + self.shares)

    order_with_respect_to = "order_proposal"

    class Meta(OrderedModel.Meta):
        verbose_name = "Order"
        verbose_name_plural = "Orders"
        indexes = [
            models.Index(fields=["order_proposal"]),
            models.Index(fields=["underlying_instrument", "value_date"]),
            models.Index(fields=["portfolio", "underlying_instrument", "value_date"]),
            models.Index(fields=["order_proposal", "underlying_instrument"]),
            # models.Index(fields=["date", "underlying_instrument"]),
        ]
        constraints = [
            models.UniqueConstraint(
                fields=["order_proposal", "underlying_instrument"],
                name="unique_order",
            ),
        ]
        # notification_email_template = "portfolio/email/trade_notification.html"

    def pre_save(self):
        self.portfolio = self.order_proposal.portfolio
        self.value_date = self.order_proposal.trade_date

        if not self.price:
            # we try to get the price if not provided directly from the underlying instrument
            if self.underlying_instrument.is_cash or self.underlying_instrument.is_cash_equivalent:
                self.price = Decimal("1")
            else:
                self.price = self.get_price()
        if not self.portfolio.only_weighting:
            estimated_shares = self.order_proposal.get_estimated_shares(
                self.weighting, self.underlying_instrument, self.price
            )
            if estimated_shares:
                self.shares = estimated_shares
        super().pre_save()

    def save(self, *args, **kwargs):
        self.pre_save()
        if not self.underlying_instrument.is_investable_universe:
            self.underlying_instrument.is_investable_universe = True
            self.underlying_instrument.save()

        if self.id:
            self.set_type()
        super().save(*args, **kwargs)

    @classmethod
    def get_type(cls, weighting, effective_weight, target_weight) -> Type:
        if weighting == 0:
            return Order.Type.NO_CHANGE
        elif weighting is not None:
            if weighting > 0:
                if abs(effective_weight) > 1e-8:
                    return Order.Type.INCREASE
                else:
                    return Order.Type.BUY
            elif weighting < 0:
                if abs(target_weight) > 1e-8:
                    return Order.Type.DECREASE
                else:
                    return Order.Type.SELL

    def set_type(self):
        self.order_type = self.get_type(self.weighting, self._previous_weight, self._target_weight)

    def get_price(self) -> Decimal:
        try:
            return self.underlying_instrument.get_price(self.value_date)
        except ValueError:
            return Decimal("0")

    def __str__(self):
        ticker = f"{self.underlying_instrument.ticker}:" if self.underlying_instrument.ticker else ""
        return f"{ticker}{self.weighting}"



from contextlib import suppress
from datetime import date, datetime

import pandas as pd
from django.db.models import Exists, F, OuterRef, Q, Sum
from django.utils.dateparse import parse_date
from django.utils.functional import cached_property
from wbcore import viewsets
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.pandas import fields as pf
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.serializers import decorator
from wbcore.utils.strings import format_number
from wbfdm.contrib.metric.viewsets.mixins import InstrumentMetricMixin
from wbfdm.models import Instrument

from wbportfolio.filters import (
    AssetPositionFilter,
    AssetPositionInstrumentFilter,
    AssetPositionPortfolioFilter,
    CashPositionPortfolioFilterSet,
    CompositionModelPortfolioPandasFilter,
)
from wbportfolio.import_export.resources.assets import AssetPositionResource
from wbportfolio.metric.backends.portfolio_base import (
    PORTFOLIO_CAPITAL_EMPLOYED,
    PORTFOLIO_EBIT,
    PORTFOLIO_LIABILITIES,
    PORTFOLIO_ROCE,
    PORTFOLIO_TOTAL_ASSETS,
)
from wbportfolio.metric.backends.portfolio_esg import PORTFOLIO_ESG_KEYS
from wbportfolio.models import (
    AssetPosition,
    InstrumentPortfolioThroughModel,
    Portfolio,
    Product,
)
from wbportfolio.serializers.assets import (
    AssetPositionAggregatedPortfolioModelSerializer,
    AssetPositionInstrumentModelSerializer,
    AssetPositionModelSerializer,
    AssetPositionPortfolioModelSerializer,
    CashPositionPortfolioModelSerializer,
)

from .configs import (
    AssetPositionButtonConfig,
    AssetPositionDisplayConfig,
    AssetPositionEndpointConfig,
    AssetPositionInstrumentButtonConfig,
    AssetPositionInstrumentDisplayConfig,
    AssetPositionInstrumentEndpointConfig,
    AssetPositionInstrumentTitleConfig,
    AssetPositionPortfolioButtonConfig,
    AssetPositionPortfolioDisplayConfig,
    AssetPositionPortfolioEndpointConfig,
    AssetPositionPortfolioTitleConfig,
    AssetPositionTitleConfig,
    CashPositionPortfolioDisplayConfig,
    CashPositionPortfolioEndpointConfig,
    CashPositionPortfolioTitleConfig,
    CompositionModelPortfolioPandasDisplayConfig,
    CompositionModelPortfolioPandasEndpointConfig,
    CompositionModelPortfolioPandasTitleConfig,
)
from .mixins import UserPortfolioRequestPermissionMixin


class AssetPositionModelViewSet(
    UserPortfolioRequestPermissionMixin,
    InternalUserPermissionMixin,
    viewsets.ReadOnlyModelViewSet,
):
    IDENTIFIER = "wbportfolio:assetposition"
    IMPORT_ALLOWED = False

    queryset = AssetPosition.objects.all()
    serializer_class = AssetPositionModelSerializer
    filterset_class = AssetPositionFilter

    ordering_fields = [
        "total_value_fx_portfolio",
        "total_value_fx_usd",
        "total_value",
        "portfolio__name",
        "portfolio_created__name",
        "underlying_quote",
        "underlying_quote_name",
        "underlying_quote_ticker",
        "underlying_quote_isin",
        "price",
        "currency__key",
        "currency_fx_rate",
        "shares",
        "date",
        "asset_valuation_date",
        "weighting",
        "market_share",
        "liquidity",
    ]
    ordering = ["-weighting"]
    search_fields = ["underlying_quote_ticker", "underlying_quote_name", "underlying_quote_isin"]

    display_config_class = AssetPositionDisplayConfig
    button_config_class = AssetPositionButtonConfig
    endpoint_config_class = AssetPositionEndpointConfig
    title_config_class = AssetPositionTitleConfig

    def get_resource_class(self):
        return AssetPositionResource

    def get_aggregates(self, queryset, paginated_queryset):
        aggregates = super().get_aggregates(queryset, paginated_queryset)
        if queryset.exists():
            total_value_fx_usd = queryset.aggregate(s=Sum(F("total_value_fx_usd")))["s"]
            weighting = queryset.aggregate(s=Sum(F("weighting")))["s"]
            aggregates.update(
                {
                    "weighting": {"Σ": format_number(weighting, decimal=8)},
                    "total_value_fx_usd": {"Σ": format_number(total_value_fx_usd)},
                }
            )
        return aggregates

    def get_queryset(self):
        if self.is_analyst:
            return (
                super()
                .get_queryset()
                .annotate(
                    underlying_quote_isin=F("underlying_quote__isin"),
                    underlying_quote_ticker=F("underlying_quote__ticker"),
                    underlying_quote_name=F("underlying_quote__name"),
                )
                .select_related(
                    "underlying_quote",
                    "currency",
                    "portfolio",
                    "exchange",
                    "portfolio_created",
                )
            )
        return AssetPosition.objects.none()


# Portfolio Viewsets


class AssetPositionPortfolioModelViewSet(InstrumentMetricMixin, AssetPositionModelViewSet):
    METRIC_KEYS = (
        PORTFOLIO_EBIT,
        PORTFOLIO_TOTAL_ASSETS,
        PORTFOLIO_LIABILITIES,
        PORTFOLIO_CAPITAL_EMPLOYED,
        PORTFOLIO_ROCE,
        *PORTFOLIO_ESG_KEYS,
    )
    METRIC_BASKET_LABEL = "portfolio"
    METRIC_INSTRUMENT_LABEL = "underlying_quote"
    METRIC_SHOW_BY_DEFAULT = False
    METRIC_SHOW_FILTERS = True

    @property
    def metric_date(self):
        if date_str := self.request.GET.get("date"):
            return parse_date(date_str)

    @property
    def metric_basket(self):
        return self.portfolio

    filterset_class = AssetPositionPortfolioFilter
    display_config_class = AssetPositionPortfolioDisplayConfig
    button_config_class = AssetPositionPortfolioButtonConfig
    title_config_class = AssetPositionPortfolioTitleConfig
    endpoint_config_class = AssetPositionPortfolioEndpointConfig

    def get_serializer_class(self):
        if self.request.GET.get("aggregate", "false") == "true":
            return AssetPositionAggregatedPortfolioModelSerializer
        return AssetPositionPortfolioModelSerializer

    def get_aggregates(self, queryset, paginated_queryset):
        if not queryset.exists():
            return {}
        weighting = queryset.aggregate(s=Sum(F("weighting")))["s"]
        total_value_fx_portfolio = queryset.aggregate(s=Sum(F("total_value_fx_portfolio")))["s"]
        aggregates = super().get_aggregates(queryset, paginated_queryset)
        aggregates["total_value_fx_portfolio"] = {"Σ": format_number(total_value_fx_portfolio)}
        aggregates["weighting"] = {"Σ": format_number(weighting, decimal=8)}
        return aggregates

    def get_queryset(self):
        if self.has_portfolio_access:
            return (
                super()
                .get_queryset()
                .filter(portfolio=self.portfolio)
                .select_related("underlying_quote", "currency", "exchange", "portfolio_created")
            )
        return AssetPosition.objects.none()


# Underlying Assets viewsets
class AssetPositionInstrumentModelViewSet(AssetPositionModelViewSet):
    filterset_class = AssetPositionInstrumentFilter
    serializer_class = AssetPositionInstrumentModelSerializer
    display_config_class = AssetPositionInstrumentDisplayConfig
    title_config_class = AssetPositionInstrumentTitleConfig
    button_config_class = AssetPositionInstrumentButtonConfig
    endpoint_config_class = AssetPositionInstrumentEndpointConfig

    def get_aggregates(self, queryset, paginated_queryset):
        queryset = queryset.filter(is_invested=True)
        if queryset.exists():
            total_value_fx_usd = queryset.aggregate(s=Sum(F("total_value_fx_usd")))["s"]
            total_shares = queryset.aggregate(s=Sum(F("shares")))["s"]
            total_market_share = queryset.aggregate(s=Sum(F("market_share")))["s"]
            return {
                "shares": {"Σ": format_number(total_shares)},
                "total_value_fx_usd": {"Σ": format_number(total_value_fx_usd)},
                "market_share": {"Σ": format_number(total_market_share)},
            }
        return {}

    def get_queryset(self):
        qs = super().get_queryset().filter(underlying_quote__in=self.instrument.get_descendants(include_self=True))
        if self.request.GET.get("filter_last_positions", "false") == "true":
            if self.instrument.assets.exists():
                qs = qs.filter(date=self.instrument.assets.latest("date").date)
        return qs


class CashPositionPortfolioPandasAPIView(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet
):
    queryset = AssetPosition.objects.all()
    display_config_class = CashPositionPortfolioDisplayConfig
    title_config_class = CashPositionPortfolioTitleConfig
    endpoint_config_class = CashPositionPortfolioEndpointConfig

    serializer_class = CashPositionPortfolioModelSerializer
    filterset_class = CashPositionPortfolioFilterSet

    search_fields = ["portfolio_name"]
    ordering_fields = ["portfolio_name", "total_value_fx_usd", "portfolio_weight"]

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="portfolio", label="ID"),
            pf.CharField(key="portfolio_name", label="Portfolio"),
            pf.FloatField(
                key="total_value_fx_usd",
                label="Total Value",
                precision=2,
                decorators=(decorator(decorator_type="text", position="right", value="$"),),
            ),
            pf.FloatField(key="portfolio_weight", label="Total portfolio value", precision=2, percent=True),
        )
    )

    def get_aggregates(self, request, df):
        if not df.empty:
            sum_total_value_fx_usd = df.total_value_fx_usd.sum()
            return {
                "total_value_fx_usd": {"Σ": format_number(sum_total_value_fx_usd)},
            }
        return {}

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame()
        if queryset.exists():
            df = pd.DataFrame(
                queryset.values(
                    "underlying_instrument__is_cash",
                    "total_value_fx_portfolio",
                    "fx_rate",
                    "portfolio__name",
                    "portfolio",
                )
            )
            df_name = (
                df[["portfolio", "portfolio__name"]]
                .groupby("portfolio")
                .agg("first")
                .rename(columns={"portfolio__name": "portfolio_name"})
            )

            df["total_value_fx_usd"] = df.total_value_fx_portfolio * df.fx_rate
            df_total = (
                df[["portfolio", "total_value_fx_usd"]]
                .groupby("portfolio")
                .sum()
                .rename(columns={"total_value_fx_usd": "total_portfolio_fx_usd"})
            )
            df = df[df["underlying_instrument__is_cash"]]
            df = df[["portfolio", "total_value_fx_usd"]].groupby("portfolio").sum()
            if not df.empty:
                df = df[df["total_value_fx_usd"] != 0]
                df = pd.concat([df, df_total, df_name], axis=1).dropna(how="any")
                df["portfolio_weight"] = df.total_value_fx_usd / df.total_portfolio_fx_usd
                return df.reset_index()
        return df

    def get_queryset(self):
        if self.is_portfolio_manager:
            if date_str := self.request.GET.get("date", None):
                val_date = datetime.strptime(date_str, "%Y-%m-%d").date()
            else:
                val_date = date.today()
            active_products = Product.active_objects.filter_active_at_date(val_date)
            return (
                super()
                .get_queryset()
                .annotate(
                    fx_rate=CurrencyFXRates.get_fx_rates_subquery(
                        "date", currency="portfolio__currency", lookup_expr="exact"
                    ),
                    is_product_portfolio=Exists(
                        InstrumentPortfolioThroughModel.objects.filter(
                            portfolio=OuterRef("portfolio"), instrument__in=active_products
                        )
                    ),
                )
                .filter(is_product_portfolio=True)
            )
        return AssetPosition.objects.none()


class CompositionModelPortfolioPandasView(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet
):
    IDENTIFIER = "wbportfolio:topdowncomposition"
    filterset_class = CompositionModelPortfolioPandasFilter
    queryset = AssetPosition.objects.all()

    display_config_class = CompositionModelPortfolioPandasDisplayConfig
    title_config_class = CompositionModelPortfolioPandasTitleConfig
    endpoint_config_class = CompositionModelPortfolioPandasEndpointConfig

    @cached_property
    def val_date(self) -> date:
        return parse_date(self.request.GET["date"])

    @cached_property
    def dependant_portfolios(self):
        return Portfolio.objects.filter(
            Q(depends_on__in=[self.portfolio]) | Q(dependent_portfolios__in=[self.portfolio])
        )

    def get_queryset(self):
        if self.has_portfolio_access:
            return super().get_queryset().filter(portfolio=self.portfolio)
        return AssetPosition.objects.none()

    def get_pandas_fields(self, request):
        fields = [
            pf.PKField(key="underlying_instrument", label="ID"),
            pf.CharField(key="underlying_instrument_repr", label="Instrument"),
        ]
        if not self.portfolio.only_weighting:
            fields.append(pf.FloatField(key=f"shares_{self.portfolio.id}", label=str(self.portfolio)))
        fields.append(pf.FloatField(key=f"weighting_{self.portfolio.id}", label=str(self.portfolio), percent=True))

        for portfolio in self.dependant_portfolios:
            if not portfolio.only_weighting:
                fields.append(pf.FloatField(key=f"shares_{portfolio.id}", label=str(portfolio)))
                fields.append(pf.FloatField(key=f"difference_{portfolio.id}", label=str(portfolio)))
            else:
                fields.append(pf.FloatField(key=f"weighting_{portfolio.id}", label=str(portfolio), percent=True))
                fields.append(pf.FloatField(key=f"difference_{portfolio.id}", label=str(portfolio), percent=True))

        return pf.PandasFields(fields=fields)

    search_fields = ["underlying_instrument_repr"]
    ordering_fields = ["underlying_instrument_repr", "model_portfolio"]

    def get_dataframe(self, request, queryset, **kwargs):
        rows = []
        for portfolio in [*self.dependant_portfolios, self.portfolio]:
            rows.extend(
                list(
                    map(
                        lambda x: {
                            "underlying_instrument": x.underlying_instrument.id,
                            "weighting": x.weighting,
                            "shares": x.shares,
                            "portfolio": portfolio.id,
                        },
                        portfolio.get_positions(self.val_date),
                    )
                )
            )
        df = pd.DataFrame(rows, columns=["underlying_instrument", "weighting", "shares", "portfolio"])
        df = df.pivot_table(
            index="underlying_instrument",
            columns=["portfolio"],
            values=["weighting", "shares"],
            aggfunc="sum",
            fill_value=0,
        )
        for portfolio in self.dependant_portfolios:
            with suppress(KeyError):
                if portfolio.only_weighting:
                    df["difference", portfolio.id] = df["weighting"][self.portfolio.id] - df["weighting"][portfolio.id]
                else:
                    df["difference", portfolio.id] = df["shares"][self.portfolio.id] - df["shares"][portfolio.id]

        df.columns = ["_".join([str(item) for item in col]) for col in df.columns.to_flat_index()]
        df = df.reset_index()

        return df

    def manipulate_dataframe(self, df):
        df["underlying_instrument_repr"] = df["underlying_instrument"].map(
            dict(Instrument.objects.filter(id__in=df["underlying_instrument"]).values_list("id", "computed_str"))
        )
        return df



from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from django.contrib.contenttypes.models import ContentType
from faker import Faker
from wbfdm.analysis.esg.enums import ESGAggregation
from wbfdm.contrib.metric.models import InstrumentMetric

from ..backends.portfolio_esg import PortfolioESGMetricBackend

fake = Faker()


@pytest.mark.django_db
class TestPortfolioESGDataloader:
    @pytest.fixture
    def backend(self, weekday):
        return PortfolioESGMetricBackend(weekday)

    @patch.object(PortfolioESGMetricBackend, "_get_metrics_for_esg_aggregation")
    def test_compute_metrics(self, mock_fct, backend, portfolio, instrument_factory, asset_position_factory):
        instruments = instrument_factory.create_batch(2)
        asset_position_factory.create(portfolio=portfolio, is_estimated=False, date=backend.val_date)
        esg_aggregation = pd.DataFrame(
            np.random.randint(1, 100, size=[2, 1]), index=[i.id for i in instruments], columns=["score"]
        )
        mock_fct.return_value = esg_aggregation
        for metric in backend.compute_metrics(portfolio):
            InstrumentMetric.update_or_create_from_metric(metric)

        p_ct = ContentType.objects.get_for_model(portfolio)

        # check that we indeed created and stored a metrics for each esg aggregation and for each instrument
        for esg in ESGAggregation:
            for instrument in instruments:
                assert (
                    InstrumentMetric.objects.get(
                        basket_content_type=p_ct.pk,
                        basket_id=portfolio.id,
                        instrument=instrument,
                        date=backend.val_date,
                        key=f"portfolio_esg_{esg.name.lower()}",
                    ).metrics["score"]
                    == esg_aggregation.loc[instrument.id, "score"]
                )

            # check if the egg aggregated accross the whole portfolio is properly stored
            assert (
                InstrumentMetric.objects.get(
                    basket_content_type=p_ct.pk,
                    basket_id=portfolio.id,
                    instrument__isnull=True,
                    date=backend.val_date,
                    key=f"portfolio_esg_{esg.name.lower()}",
                ).metrics["score"]
                == esg_aggregation.sum(axis=0)["score"]
            )

        #
        # # basic checks
        # assert res[0].basket_id == portfolio.id
        # assert res[0].basket_content_type_id ==
        # assert res[0].key ==
        # assert res[0].metrics ==
        # assert res[0].date ==
        # assert res[0].instrument_id ==
        # assert res[0].dependency_metrics ==



from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

NEGATIVE_RED_FORMATTING = [
    dp.FormattingRule(
        style={
            "color": WBColor.RED_DARK.value,
            "fontWeight": "bold",
        },
        condition=("<", 0),
    )
]

SHARE_FORMATTING = dp.Formatting(
    column="shares",
    formatting_rules=[
        dp.FormattingRule(
            style={
                "color": WBColor.RED_DARK.value,
                # "fontWeight": "bold",
            },
            condition=("<", 0),
        )
    ],
)


class TradeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="transaction_subtype", label="Type", width=Unit.PIXEL(100)),
                dp.Field(key="transaction_date", label="Trade Date", width=Unit.PIXEL(150)),
                dp.Field(key="underlying_instrument", label="Instrument", width=Unit.PIXEL(250)),
                dp.Field(
                    label="Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="shares", label="Shares", width=Unit.PIXEL(100)),
                        dp.Field(key="price", label="Price", width=Unit.PIXEL(100)),
                        dp.Field(key="total_value", label="Value", width=Unit.PIXEL(125)),
                        dp.Field(key="total_value_usd", label="Value ($)", width=Unit.PIXEL(125), show="open"),
                    ],
                ),
                dp.Field(
                    label="Bank",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="bank", label="Counterparty", width=Unit.PIXEL(150)),
                        dp.Field(key="custodian", label="Custodian", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="register", label="Register", width=Unit.PIXEL(200), show="open"),
                    ],
                ),
                dp.Field(
                    label="Information",
                    open_by_default=False,
                    key=None,
                    children=list(
                        filter(
                            lambda x: not (x.key == "portfolio" and "portfolio_id" in self.view.kwargs),
                            [
                                dp.Field(key="currency", label="Currency", width=Unit.PIXEL(100)),
                                dp.Field(key="portfolio", label="Portfolio", width=Unit.PIXEL(250), show="open"),
                                dp.Field(
                                    key="marked_for_deletion",
                                    label="Marked for Deletion",
                                    width=Unit.PIXEL(150),
                                    show="open",
                                ),
                                dp.Field(key="pending", label="Pending", width=Unit.PIXEL(150), show="open"),
                            ],
                        )
                    ),
                ),
            ],
            formatting=[SHARE_FORMATTING],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["transaction_date", "value_date", "."],
                [repeat_field(3, "underlying_instrument")],
                ["shares", "price", "bank"],
                ["external_id", "marked_for_deletion", "register"],
                [repeat_field(3, "comment")],
            ]
        )


class SubscriptionRedemptionDisplayConfig(TradeDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="transaction_subtype", label="Type", width=Unit.PIXEL(100)),
                dp.Field(key="transaction_date", label="Trade Date", width=Unit.PIXEL(150)),
                dp.Field(key="underlying_instrument", label="Product", width=Unit.PIXEL(250)),
                dp.Field(
                    label="Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="shares", label="Shares", width=Unit.PIXEL(100)),
                        dp.Field(key="price", label="Price", width=Unit.PIXEL(100)),
                        dp.Field(key="total_value", label="Value", width=Unit.PIXEL(125)),
                        dp.Field(key="total_value_usd", label="Value ($)", width=Unit.PIXEL(125), show="open"),
                    ],
                ),
                dp.Field(
                    label="Customer",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="claimed_shares", label="Claimed Shares", width=Unit.PIXEL(150)),
                        dp.Field(key="claims", label="Claims", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="comment", label="Comment", width=Unit.PIXEL(200)),
                    ],
                ),
                dp.Field(
                    label="Bank",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="bank", label="Bank", width=Unit.PIXEL(150)),
                        dp.Field(key="custodian", label="Custodian", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="register", label="Register", width=Unit.PIXEL(200), show="open"),
                    ],
                ),
                dp.Field(
                    label="Information",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="currency", label="Currency", width=Unit.PIXEL(100), show="open"),
                        dp.Field(key="portfolio", label="Portfolio", width=Unit.PIXEL(250), show="open"),
                        dp.Field(key="marked_for_deletion", label="Marked for Deletion", width=Unit.PIXEL(150)),
                        dp.Field(key="pending", label="Pending", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="marked_as_internal", label="Internal", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="internal_trade", label="Internal Trade", width=Unit.PIXEL(150), show="open"),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    key="completely_claimed",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label="Completely Claimed",
                            value=True,
                        )
                    ],
                ),
                dp.Legend(
                    key="completely_claimed_if_approved",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label="Completely Claimed if approved",
                            value=True,
                        )
                    ],
                ),
                dp.Legend(
                    key="pending",
                    items=[
                        dp.LegendItem(icon=WBIcon.FOLDERS_ADD.icon, label="Pending", value=True),
                    ],
                ),
                dp.Legend(
                    key="marked_for_deletion",
                    items=[
                        dp.LegendItem(icon=WBIcon.DELETE.icon, label="To Be Deleted", value=True),
                    ],
                ),
            ],
            formatting=[
                SHARE_FORMATTING,
                dp.Formatting(
                    column="completely_claimed",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", True),
                        )
                    ],
                ),
                dp.Formatting(
                    column="completely_claimed_if_approved",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", True),
                        )
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["transaction_date", "value_date", "."],
                [repeat_field(3, "underlying_instrument")],
                ["shares", "price", "bank"],
                ["external_id", "marked_for_deletion", "register"],
                [repeat_field(3, "comment")],
                ["marked_as_internal", repeat_field(2, "internal_trade")],
            ]
        )


class TradePortfolioDisplayConfig(TradeDisplayConfig):
    pass



# Generated by Django 5.0.3 on 2024-04-23 09:36

from django.db import migrations, models


def handle_existing_duplicates(apps, schema_editor):
    Fees = apps.get_model("wbportfolio", "Fees")

    # we delete any fees with non unique lookup with same total value, this is safe
    for row in (
        Fees.objects.values("linked_product", "fee_date", "transaction_subtype", "calculated", "total_value")
        .annotate(c=models.Count("*"))
        .filter(c__gt=1)
    ):
        row.pop("c")
        for f in Fees.objects.filter(**row)[1:]:
            f.delete()
    # we dinstiguly delete duplicates only for calculated fees and keep the last inserted
    for row in (
        Fees.objects.filter(calculated=True)
        .values("linked_product", "fee_date", "transaction_subtype", "calculated")
        .annotate(c=models.Count("*"))
        .filter(c__gt=1)
    ):
        row.pop("c")
        for f in Fees.objects.filter(**row).order_by("-id")[1:]:
            f.delete()

    # we shouldn't have any duplicates (particularly non calculated fees) at this point


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0058_pmsinstrument"),
    ]

    operations = [
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL IMMEDIATE;",
        ),
        migrations.RunPython(handle_existing_duplicates),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL DEFERRED;",
        ),
        migrations.AddConstraint(
            model_name="fees",
            constraint=models.UniqueConstraint(
                fields=("product", "fee_date", "transaction_subtype", "calculated"), name="unique_fees"
            ),
        ),
    ]



from wbcore import viewsets

from wbportfolio.models.portfolio_swing_pricings import PortfolioSwingPricing
from wbportfolio.serializers import PortfolioSwingPricingModelSerializer


class PortfolioSwingPricingModelViewSet(viewsets.ModelViewSet):
    queryset = PortfolioSwingPricing.objects.all()
    serializer_class = PortfolioSwingPricingModelSerializer



from wbcore.metadata.configs.titles import TitleViewConfig


class PerformancePandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Product Performances & AuM"


class PerformanceComparisonTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Product Performances Summary"


class ProductPerformanceNetNewMoneyTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Product Performance New Money"



from django.contrib import admin
from django_better_admin_arrayfield.admin.mixins import DynamicArrayMixin

from wbportfolio.models import Custodian


@admin.register(Custodian)
class CustodianModelAdmin(admin.ModelAdmin, DynamicArrayMixin):
    search_fields = ("name", "mapping")
    list_display = ("name", "mapping", "company")



from django.contrib import admin

from wbportfolio.models import OrderProposal

from .orders import OrderTabularInline


@admin.register(OrderProposal)
class OrderProposalAdmin(admin.ModelAdmin):
    search_fields = ["portfolio__name", "comment"]

    list_display = ("portfolio", "rebalancing_model", "trade_date", "status")
    autocomplete_fields = ["portfolio", "rebalancing_model"]
    inlines = [OrderTabularInline]



import datetime as dt
from collections import defaultdict
from contextlib import suppress
from decimal import Decimal

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from django.db.models import F, Prefetch
from django.utils.functional import cached_property
from plotly.subplots import make_subplots
from rest_framework.request import Request
from wbcore import viewsets
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.filters import DjangoFilterBackend
from wbcore.utils.date import get_date_interval_from_request
from wbcore.utils.figures import (
    get_default_timeserie_figure,
    get_hovertemplate_timeserie,
)
from wbcore.utils.strings import format_number
from wbfdm.models import (
    Classification,
    ClassificationGroup,
    Instrument,
    InstrumentClassificationThroughModel,
    InstrumentType,
)

from wbportfolio.filters.assets import (
    AssetPositionUnderlyingInstrumentChartFilter,
    CompositionContributionChartFilter,
    ContributionChartFilter,
    DistributionFilter,
)
from wbportfolio.models import (
    AssetPosition,
    AssetPositionGroupBy,
    Portfolio,
    PortfolioRole,
)

from ..configs.buttons.assets import (
    DistributionChartButtonConfig,
    DistributionTableButtonConfig,
)
from ..configs.display.assets import DistributionTableDisplayConfig
from ..configs.endpoints.assets import (
    AssetPositionUnderlyingInstrumentChartEndpointConfig,
    ContributorPortfolioChartEndpointConfig,
    DistributionChartEndpointConfig,
    DistributionTableEndpointConfig,
)
from ..configs.titles.assets import (
    AssetPositionUnderlyingInstrumentChartTitleConfig,
    ContributorPortfolioChartTitleConfig,
    DistributionChartTitleConfig,
    DistributionTableTitleConfig,
)
from ..mixins import UserPortfolioRequestPermissionMixin


class AbstractDistributionMixin(UserPortfolioRequestPermissionMixin):
    queryset = AssetPosition.objects.all()
    filterset_class = DistributionFilter
    filter_backends = (DjangoFilterBackend,)
    request: Request

    @cached_property
    def group_by(self) -> AssetPositionGroupBy:
        try:
            return AssetPositionGroupBy(self.request.GET.get("group_by", "classification"))
        except ValueError:
            return AssetPositionGroupBy.INDUSTRY

    @cached_property
    def val_date(self) -> dt.date:
        if validity_date_repr := self.request.GET.get("date"):
            val_date = dt.datetime.strptime(validity_date_repr, "%Y-%m-%d")
        else:
            val_date = dt.date.today()
        return val_date

    @cached_property
    def classification_group(self):
        try:
            return ClassificationGroup.objects.get(id=self.request.GET.get("group_by_classification_group"))
        except ClassificationGroup.DoesNotExist:
            return ClassificationGroup.objects.get(is_primary=True)

    @cached_property
    def classification_height(self) -> int:
        return int(self.request.GET.get("group_by_classification_height", "0"))

    @cached_property
    def columns_map(self) -> dict:
        if self.group_by == AssetPositionGroupBy.INDUSTRY:
            columns = {}
            level_representations = self.classification_group.get_levels_representation()
            for key, label in zip(
                reversed(self.classification_group.get_fields_names(sep="_")), reversed(level_representations[1:])
            ):
                columns[key] = label
            columns["label"] = "Classification"
            columns["equity"] = "Instrument"
        else:
            columns = {"label": "Label"}
        return columns

    def get_queryset(self):
        profile = self.request.user.profile
        if PortfolioRole.is_analyst(profile, portfolio=self.portfolio) or profile.is_internal:
            return AssetPosition.objects.filter(portfolio=self.portfolio)
        return AssetPosition.objects.none()

    def get_dataframe(self, request, queryset, **kwargs) -> pd.DataFrame:
        instruments = defaultdict(Decimal)
        for asset in self.portfolio.get_positions(self.val_date):
            if self.group_by != AssetPositionGroupBy.INDUSTRY:
                group_field = getattr(asset.underlying_instrument, self.group_by.value)
            else:
                group_field = asset.underlying_instrument.get_root()
            group_field = getattr(group_field, "id", group_field)
            instruments[group_field] += asset.weighting
        df = pd.DataFrame.from_dict(instruments, orient="index", columns=["weighting"])
        if self.group_by == AssetPositionGroupBy.INDUSTRY:
            classifications = InstrumentClassificationThroughModel.objects.filter(
                instrument__in=instruments.keys(), classification__group=self.classification_group
            )
            field_names = {
                field_name.replace("__", "_"): F(f"classification__{field_name}__name")
                for field_name in self.classification_group.get_fields_names()
            }
            classifications = (
                classifications.annotate(**field_names)
                .select_related(
                    *[f"classification__{field_name}" for field_name in self.classification_group.get_fields_names()]
                )
                .prefetch_related(
                    "tags",
                    Prefetch("instrument", queryset=Instrument.objects.filter(classifications_through__isnull=False)),
                )
            )
            df_classification = (
                pd.DataFrame(
                    classifications.values_list("instrument", "classification", *field_names.keys()),
                    columns=["id", "classification", *field_names.keys()],
                )
                .groupby("id")
                .first()
            )
            df = pd.concat([df, df_classification], axis=1)
        if df.weighting.sum():  # normalize
            df.weighting /= df.weighting.sum()
        return df.reset_index(names="id")

    def manipulate_dataframe(self, df):
        if self.group_by == AssetPositionGroupBy.INDUSTRY:
            if not PortfolioRole.is_analyst(self.request.user.profile, portfolio=self.portfolio):
                df["equity"] = ""
            else:
                df["equity"] = df["id"].map(
                    dict(Instrument.objects.filter(id__in=df["id"]).values_list("id", "computed_str"))
                )
            df["label"] = df["classification"].map(
                dict(Classification.objects.filter(id__in=df["classification"].dropna()).values_list("id", "name"))
            )
        elif self.group_by == AssetPositionGroupBy.CASH:
            df.loc[df["id"] == True, "label"] = "Cash"
            df.loc[df["id"] == False, "label"] = "Non-Cash"
        elif self.group_by == AssetPositionGroupBy.COUNTRY:
            df["label"] = df["id"].map(dict(Geography.objects.filter(id__in=df["id"]).values_list("id", "name")))
        elif self.group_by == AssetPositionGroupBy.CURRENCY:
            currencies = dict(map(lambda o: (o.id, str(o)), Currency.objects.filter(id__in=df["id"])))
            df["label"] = df["id"].map(currencies)
        elif self.group_by == AssetPositionGroupBy.INSTRUMENT_TYPE:
            df["label"] = df["id"].map(
                dict(InstrumentType.objects.filter(id__in=df["id"]).values_list("id", "short_name"))
            )
        df.sort_values(by="weighting", ascending=False, inplace=True)
        return df


class DistributionChartViewSet(AbstractDistributionMixin, viewsets.ChartViewSet):
    title_config_class = DistributionChartTitleConfig
    endpoint_config_class = DistributionChartEndpointConfig
    button_config_class = DistributionChartButtonConfig

    @staticmethod
    def pie_chart(df: pd.DataFrame) -> go.Figure:
        fig = go.Figure()
        if not df.empty:
            fig.add_pie(
                labels=df.index,
                values=df.weighting.mul(100),
                marker=dict(colors=px.colors.qualitative.Plotly[: df.shape[0]], line=dict(color="#000000", width=2)),
                hovertemplate="<b>%{label}</b><extra></extra>",
            )
        return fig

    def get_plotly(self, queryset):
        fig = go.Figure()
        df = self.manipulate_dataframe(self.get_dataframe(self.request, queryset))
        df = df.dropna(how="any")
        if not df.empty:
            levels = list(self.columns_map.keys())
            fig = px.sunburst(
                df,
                path=levels,
                values="weighting",
                hover_data={"weighting": ":.2%"},
            )
        fig.update_traces(hovertemplate="<b>%{label}</b><br>Weight = %{customdata:.3p}")
        return fig


class DistributionTableViewSet(AbstractDistributionMixin, ExportPandasAPIViewSet):
    endpoint_config_class = DistributionTableEndpointConfig
    display_config_class = DistributionTableDisplayConfig
    title_config_class = DistributionTableTitleConfig
    button_config_class = DistributionTableButtonConfig

    def get_pandas_fields(self, request):
        fields = [
            pf.PKField(key="id", label="id"),
            pf.FloatField(key="weighting", label="Weight", precision=2, percent=True),
        ]
        for key, label in self.columns_map.items():
            fields.append(pf.CharField(key=key, label=label))
        return pf.PandasFields(fields=fields)

    def get_aggregates(self, request, df):
        return {"weighting": {"Σ": format_number(df["weighting"].sum())}}


# ##### CHART VIEWS #####


class ContributorPortfolioChartView(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
    filterset_class = ContributionChartFilter
    filter_backends = (DjangoFilterBackend,)
    IDENTIFIER = "wbportfolio:portfolio-contributor"
    queryset = AssetPosition.objects.all()

    title_config_class = ContributorPortfolioChartTitleConfig
    endpoint_config_class = ContributorPortfolioChartEndpointConfig

    ROW_HEIGHT: int = 20

    @property
    def min_height(self):
        if hasattr(self, "nb_rows"):
            return self.nb_rows * self.ROW_HEIGHT
        return "300px"

    @cached_property
    def hedged_currency(self) -> Currency | None:
        if "hedged_currency" in self.request.GET:
            with suppress(Currency.DoesNotExist):
                return Currency.objects.get(pk=self.request.GET["hedged_currency"])

    @cached_property
    def show_lookthrough(self) -> bool:
        return self.portfolio.is_composition and self.request.GET.get("show_lookthrough", "false").lower() == "true"

    def get_filterset_class(self, request):
        if self.portfolio.is_composition:
            return CompositionContributionChartFilter
        return ContributionChartFilter

    def get_plotly(self, queryset):
        fig = go.Figure()
        data = []
        if self.show_lookthrough:
            d1, d2 = get_date_interval_from_request(self.request)
            for _d in pd.date_range(d1, d2):
                for pos in self.portfolio.get_lookthrough_positions(_d.date()):
                    data.append(
                        [
                            pos.date,
                            pos.initial_price,
                            pos.initial_currency_fx_rate,
                            pos.underlying_instrument_id,
                            pos.weighting,
                        ]
                    )
        else:
            data = queryset.annotate_hedged_currency_fx_rate(self.hedged_currency).values_list(
                "date", "price", "hedged_currency_fx_rate", "underlying_instrument", "weighting"
            )
        df = Portfolio.get_contribution_df(data).rename(columns={"group_key": "underlying_instrument"})
        if not df.empty:
            df = df[["contribution_total", "contribution_forex", "underlying_instrument"]].sort_values(
                by="contribution_total", ascending=True
            )

            df["instrument_id"] = df.underlying_instrument.map(
                dict(Instrument.objects.filter(id__in=df["underlying_instrument"]).values_list("id", "name_repr"))
            )
            df_forex = df[["instrument_id", "contribution_forex"]]
            df_forex = df_forex[df_forex.contribution_forex != 0]

            contribution_equity = df.contribution_total - df.contribution_forex

            text_forex = df_forex.contribution_forex.apply(lambda x: f"{x:,.2%}")
            text_equity = contribution_equity.apply(lambda x: f"{x:,.2%}")
            setattr(self, "nb_rows", df.shape[0])
            fig.add_trace(
                go.Bar(
                    y=df.instrument_id,
                    x=contribution_equity,
                    name="Contribution Equity",
                    orientation="h",
                    marker=dict(
                        color="rgba(247,110,91,0.6)",
                        line=dict(color="rgb(247,110,91,1.0)", width=2),
                    ),
                    text=text_equity.values,
                    textposition="auto",
                )
            )
            fig.add_trace(
                go.Bar(
                    y=df_forex.instrument_id,
                    x=df_forex.contribution_forex,
                    name="Contribution Forex",
                    orientation="h",
                    marker=dict(
                        color="rgba(58, 71, 80, 0.6)",
                        line=dict(color="rgba(58, 71, 80, 1.0)", width=2),
                    ),
                    text=text_forex.values,
                    textposition="outside",
                )
            )
            fig.update_layout(
                barmode="relative",
                xaxis=dict(showgrid=False, showline=False, zeroline=False, tickformat=".2%"),
                yaxis=dict(showgrid=False, showline=False, zeroline=False, tickmode="linear"),
                margin=dict(b=0, r=20, l=20, t=0, pad=20),
                paper_bgcolor="rgba(0,0,0,0)",
                plot_bgcolor="rgba(0,0,0,0)",
                font=dict(family="roboto", size=12, color="black"),
                bargap=0.3,
            )
            # fig = get_horizontal_barplot(df, x_label="contribution_total", y_label="name")
        return fig

    def parse_figure_dict(self, figure_dict: dict[str, any]) -> dict[str, any]:
        figure_dict = super().parse_figure_dict(figure_dict)
        figure_dict["style"]["minHeight"] = self.min_height
        return figure_dict

    def get_queryset(self):
        if self.has_portfolio_access:
            return super().get_queryset().filter(portfolio=self.portfolio)
        return AssetPosition.objects.none()


class AssetPositionUnderlyingInstrumentChartViewSet(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
    IDENTIFIER = "wbportfolio:assetpositionchart"

    queryset = AssetPosition.objects.all()

    title_config_class = AssetPositionUnderlyingInstrumentChartTitleConfig
    endpoint_config_class = AssetPositionUnderlyingInstrumentChartEndpointConfig
    filterset_class = AssetPositionUnderlyingInstrumentChartFilter

    def get_queryset(self):
        return AssetPosition.objects.filter(underlying_quote__in=self.instrument.get_descendants(include_self=True))

    def get_plotly(self, queryset):
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        fig = get_default_timeserie_figure(fig)
        if queryset.exists():
            df_weight = pd.DataFrame(queryset.values("date", "weighting", "portfolio__name"))
            df_weight = df_weight.where(pd.notnull(df_weight), 0)
            df_weight = df_weight.groupby(["date", "portfolio__name"]).sum().reset_index()
            min_date = df_weight["date"].min()
            max_date = df_weight["date"].max()

            df_price = (
                pd.DataFrame(
                    self.instrument.prices.filter_only_valid_prices()
                    .annotate_base_data()
                    .filter(date__gte=min_date, date__lte=max_date)
                    .values_list("date", "net_value_usd"),
                    columns=["date", "price_fx_usd"],
                )
                .set_index("date")
                .sort_index()
            )

            fig.add_trace(
                go.Scatter(
                    x=df_price.index, y=df_price.price_fx_usd, mode="lines", marker_color="green", name="Price"
                ),
                secondary_y=False,
            )

            df_weight = pd.DataFrame(queryset.values("date", "weighting", "portfolio__name"))
            df_weight = df_weight.where(pd.notnull(df_weight), 0)
            df_weight = df_weight.groupby(["date", "portfolio__name"]).sum().reset_index()
            for portfolio_name, df_tmp in df_weight.groupby("portfolio__name"):
                fig.add_trace(
                    go.Scatter(
                        x=df_tmp.date,
                        y=df_tmp.weighting,
                        hovertemplate=get_hovertemplate_timeserie(is_percent=True),
                        mode="lines",
                        name=f"Allocation: {portfolio_name}",
                    ),
                    secondary_y=True,
                )

            # Set x-axis title
            fig.update_xaxes(title_text="Date")
            # Set y-axes titles
            fig.update_yaxes(
                title_text="<b>Price</b>",
                secondary_y=False,
                titlefont=dict(color="green"),
                tickfont=dict(color="green"),
            )
            fig.update_yaxes(
                title_text="<b>Portfolio Allocation (%)</b>",
                secondary_y=True,
                titlefont=dict(color="blue"),
                tickfont=dict(color="blue"),
            )

        return fig



import factory
from faker import Faker
from pandas.tseries.offsets import BDay

from wbportfolio.models import Adjustment

fake = Faker()


class AdjustmentFactory(factory.django.DjangoModelFactory):
    date = factory.LazyAttribute(lambda o: (fake.future_date() + BDay(0)).date())
    instrument = factory.SubFactory("wbfdm.factories.instruments.InstrumentFactory")
    factor = factory.Faker("pydecimal", min_value=2, max_value=10)
    last_handler = factory.SubFactory("wbcore.contrib.directory.factories.entries.PersonFactory")
    status = Adjustment.Status.PENDING

    class Meta:
        model = Adjustment
